
;;; =============================================================== ;;;
;;; 
;;; Copyright (C) 2004 Daniel Mayer (post.dm@gmx.at) 
;;; 
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or (at your option) any later version.
;;; 
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;; 
;;; =============================================================== ;;;

;;; Gentlecoord 1.0
;;; Last Revision: 16.6.04


; For running the following MIDI examples the OS / Lisp / CM combination 
; must have been installed properly, optionally including the interfaces for realtime MIDI output
; before the load of this interface. See the readme and installation files. 
; Without realtime MIDI the data may be written to a MIDI file, specifying 
; the filename by the :to keyword of 'pat-events' (i.e. :to "test.midi"). 

; The pattern expansion with the 'pat' macro is used to let the same code 
; run under cm-2.4.2 and cm-2.3.4 (see the reference). Of course everywhere 
; CM's macro 'new' may be used directly. 'install-setup' is only to be called once 
; and then for every use of a different setup or for refreshing after changes.
; If :max-time is not specified within 'pat-events', the process will run for one minute.
; CM syntax is not explicitely explained. See the CM documenation for the 
; use of rhythmic symbols, pattern options etc.


;;;==========================================;;;


; Ex. 1

(install-setup #!temp)

(setf rp 
      (list 
       (pat random (x weight 20) e t #(s) #(w))
       (pat random (t weight 3) e w #(h))))

(setf mp 
      (list 
       (pat pval (between 60 95))
       (pat pval (between 25 48))))

(pat-events rp mp  '(0 10)
            :tempo 90
            :no-reps '(3 1)  
            :no-8 t
            :no-8-steps t)

; A simple example in two voices, using the instruments nr. 0 and 10 of
; the current setup. Patterns are defined for rhythm and absolute pitch
; of each voice. The 'pval' ('pat pval' produces a cycle with a pval inside)
; causes the form to be evaluated at every call of 'next'.
; The rhythm is given by a pattern with symbolic durations. Rests are
; put into vectors.
; The options :no-8 and :no-8-steps generally forbid octaves between the voices
; and octave steps within one voice. :no-reps '(3 1) forces voice number 0 to
; avoid pitch repetitions within an area of 3 events. For voice number 1 
; only the last pitch is checked. After calculation number and type
; of situations, which lead to a forced nexting of the pitch pattern, are printed.


;;;==========================================;;;


; Ex. 2

(install-setup #!qt)

(setf rhy 
      (pat heap 
           (pat cycle x :for 50)
           (pat heap q h #(e) :for 7)
           (pat cycle x #(t.) x #(x.) x #(t.) 
                :for (pat random 20 40 70))
           :name 'r))

(setf mel (pat random 1.5 -1.5 2.5 -2.5 5.5 -5.5))


(setf dyn 
      (pat case             
           :on 'r
           :by ((q h) (pat cycle #(30 120) #(120 30))
                t (pat range :from 80 downto 65 to 105
                       stepping (pat random -5 5)))))
               
(pat-events rhy mel  2
            :dyn dyn
            :tempo 50
            :step-start 75
            :lb 45
            :ub 95
            :no-reps 2
            :print t)

; An example in a quartertone setup. Pitch is given by a step pattern. This
; modus is taken because :step-start is set to a non-nil value within 'pat-events'. 
; Lower and upper bound are set, if the absolute pitch exceeds them,
; a next elemnent of the pattern is read.
; The dynamics are given by a 'case'-pattern. It refers to the rhythm pattern and switches
; between a random pattern with crescendo / decrescendo for a current half or quarter
; note or a random walk for any other duration (catching condition with t).
; Within one voice the rhythm pattern is nexted before all other event controlled 
; options, given in input order of 'pat-events'  (here: :dyn) and the pitch pattern at last. 
; This evaluation order has to be remembered when building a conditional pattern
; in this way. Both subpatterns of the case pattern are only instantiated once at
; the beginning as :type is not specified and :once is its default.


;;;==========================================;;;


; Order of evaluation of several voices:


; Ex. 3

(install-setup #!temp)

(setf rhy
      (list
       (pat line #(s) 
            (pat pval (pv 'timer)))
       (pat line #(s)
            (pat pval (pv 'timer)))
       (pat line s :name 'timer)))

(setf mel
      (list
       (pat pval (+ (pv 'bass) 31))
       (pat pval (+ (pv 'bass) 28))
       (pat pval (between 30 50) :name 'bass)))

(pat-events rhy mel  '(0 1 10)
            :level '(90 90 120)
            :tempo 60
;;;         :voice-order '(2 1 0)
            )


; When patterns of voices refer to patterns of other voices the order of evaluation
; becomes important. Pitch and rhythm patterns of the upper two voices refer at
; the "same" times to the bass. As the parallel processes are slightly time-shifted
; in list-order they refer to the last bass event. If the first two line patterns
; would not start with a rest, they would ask for a last pattern value of 'timer', 
; although 'timer' hadn't been called with 'next' before. Commenting out the 
; option :voice-order changes the order of the time-shift and the upper voices 
; refer to the current bass event, generating a sequence of major thirds.


;;;==========================================;;;


; Ex. 4

(install-setup #!temp)


(setf rhy 
      (list 
       (pat cycle (pat random #(q) #(h)) (pat heap e e q h :for 2))
       (pat cycle (pat random #(q) #(h)) (pat heap e e q h :for 2))
       (pat cycle (pat random #(q) #(h)) (pat heap e e q h :for 2))))

(setf mel 
      (let ((row (pat cycle 0 7 6 10 9 3 8 2 1 5 4 11)))
        (list
         (pat pval (+ (next row) (pick 30 42)))
         (pat pval (+ (next row) (pick 54 66)))
         (pat pval (+ (next row) (pick 78 90))))))

(setf dyn 
      (list 
       (pat cycle #(10 110) #(110 10))
       (pat cycle #(10 110) #(110 10))
       (pat cycle #(10 110) #(110 10))))
 
(pat-events rhy mel  '(1 2 0) :dyn dyn :lb 30 :ub 102)
 

; Pitch reference to a pattern in a context of dynamic extent. Every time
; the pitch pattern of one voice is to be nexted, it takes the next element of 'row'.


;;;==========================================;;;


; Ex. 5

(install-setup #!temp)
          
(setf r 
      (list 
       (pat random 
            (pat cycle t #(t))
            (#(s) weight 0.1))
       (pat cycle 
            (pat random #(w) #(q) #(h) :for 1)
            (pat random (e weight 2) q. w :for 1))))

(setf m 
      (list
       (pat pval (between 25 85)
            :returning (lambda (x) (loop repeat 2 collect (+ x (between -14 14)))))
       (pat chord (pat pval (between 25 85) :for 3))))
         
(setf dyn 
      (list
       (pat range :from 70 downto 40 to 80 stepping (pat random 5 -5))
       (pat cycle #(30 100) #(80 20))))


(pat-events r m  '(10 5)
            :dyn dyn
            :tempo 120
            :no-8 '((0) (1))
            :no-1 t
            :no-reps t)

; An example with chords within voices. The grouping list argument for :no-8 means, that
; octaves are avoided only within the chords of a voice. Primes are avoided within the
; chords and between the voices. 


;;;==========================================;;;


; Ex. 6

(install-setup #!temp)

(setf rhy 
      (list
       (pat cycle #(t)
            (pat heap t t t #(t)) )
       (pat cycle 
            (pat heap t t t #(t)) 
            (pat random #(q) #(e) #(s) :for 1) )))

(setf mel
      (loop repeat 2 collect
            (pat cycle
                 (pat filter (pat heap 1 -1 2 -2) :for 2)
                 (pat filter (pat heap 6 -6 7 -7)))))
           

(setf step-start
      (list
       (pat pval (between 30 50) :for 3)
       (pat pval (between 65 90) :for 3)))
       
(setf dyn 
      (list
       (pat random 100 80 70)
       (pat random 100 80 70)))
       
(pat-events rhy mel '(10 1) 
            :dyn dyn
            :level '(100 120)
            :step-start step-start
            :tempo 65
            :no-8 t
            :no-1 t
            :no-8-steps 3
            :no-reps t)



; Filter patterns are used to preserve a motivic shape under
; restricting conditions. Both pitch patterns are defined by steps,
; two short steps are followed by a larger one. The 3-tone groups
; start from every entry of a new period of the step-start patterns.
; The filter in the pitch pattern causes the wrapped heap to
; be nexted, if a conflict situation occurs. 
; If the elements of the heap could not solve
; the situation, the algorithm would fail. Then one could 
; - after having a look at the error history - just try
; again after a break with the same or less restrictive conditions.
; Dynamics are controlled by patterns and :level.


;;;==========================================;;;


; Ex. 7

(install-setup #!temp)

(setf rhy 
      (pat heap
           (pat funcall 
                (lambda () (explsegs (between 15 30) (between 3 15) 8)))
           (pat pval (between 1.5 3) 
                :for (pat pval (between 3 10)))
           (pat funcall
                (lambda () (nreverse (explsegs (between 15 30) (between 3 15) 8)))))
      mel
      (pat chord (pat pval (between -10 11) :for 2))
      dyn 
      (pat heap
           (pat range :from 70 to 110 by (pat random 8 10 16))
           (pat range :from 110 downto 70 by (pat random 8 10 16))))

(pat-events rhy mel 10
            :dyn dyn
            :abs-times t
            :step-start '((70 73))
            :lb 25
            :ub 95
            :no-reps 2
            :no-8-steps 2
            :no-8 t
            :no-1 t
            :max-step 10
            :print t)


; Example with absolute times. A heap chooses from exponential accelerandi and
; ritartandi, calculated with CM's explsegs function, and a random "fermata" section.
; There is just one voice, producing pitch chords by step chords 
; and a step-start of one chord.  


;;;==========================================;;;


; Ex. 8

(install-setup #!temp)

(setf harmony
      (pat pval 
           (next (pat heap 0 1 2 3 4 5 6 7 8 9 10 11) 5)
           :name 'scale)
      harmony-rhy
      (pat random w b))

(setf rhy 
      (list 
       (pat cycle t #(t))
       (pat cycle t #(t))
       (pat cycle t #(t))))

(setf mel 
      (list 
       (pat pval (+ (pickl (pv 'scale))  (pick 60 72 84)))
       (pat pval (+ (pickl (pv 'scale))  (pick 48 60 72)))
       (pat pval (+ (pickl (pv 'scale))  (pick 32 48 60)))))
                                                           
(setf dyn 
      (list 
       (pat range :from 85 downto 30 to 105
            stepping (pat random -5 5))
       (pat range :from 85 downto 30 to 105
            stepping (pat random -5 5))
       (pat range :from 85 downto 30 to 105
            stepping (pat random -5 5))))
  
    
(pat-events rhy mel  '(1 2 10) 
            :v-data-pats harmony
            :v-rhythm-pats harmony-rhy
            :dyn dyn 
            :no-reps t
            :no-8 t
            :no-1 t
            :tempo '(180 135 90))

; The v-data pattern 'harmony' gives out pitch class scales of five elements. 
; They last for the times, given by the items of the v-rhythm pattern
; 'harmony-rhy', whole and brevis. :v-tempo and :v-beat default to 60 and 0.25.
; The other voices have different tempos for equally defined rhythm patterns.
; A generally defined tempo would overrule :v-tempo.


;;;==========================================;;;


; Two variations, related to the last example, but using the :pitch-scale option:
; Vectors of pitches are defined as pitch scales.
; Patterns for rhythm, dynamics and steps are equally defined for both voices.


(setf scale-1a
      (coerce (loop for i from 30 to 90 by 1 collect i) 'array)
      scale-1b
      (coerce (loop for i from 30 to 90 by 5 collect i) 'array)
      scale-2
      (coerce (loop for i from 31 to 91 by 2 collect i) 'array))


; Ex. 9a

(install-setup #!temp)

(pat-events 
 (loop repeat 2 collect
       (pat random
            (pat cycle #(t) :for (pat pval (between 5 20)))
            (pat cycle t #(t) :for (pat pval (between 5 20)))
            (pat cycle t. #(t.) :for (pat pval (between 5 20)))))
 (loop repeat 2 collect (pat random 1 -1 2 -2))
 '(10 5)
 :dyn (loop repeat 2 collect (pat heap 70 90 110)) 
 :v-rhythm-pats (pat cycle 2)
 :v-data-pats (pat cycle :of `(,scale-1a ,scale-1b) :name 'two-scales)
 :abs-times t
 :max-time 30
 :tempo 180
 :step-start '(50 60)
 :pitch-scale `(two-scales ,scale-2)
 :no-8 t
 :no-1 t)


; In Ex. 9a the v-data pattern named 'two-scales' switches between the scales of
; semitones and fourths every two seconds (:abs-times is set t). The :pitch-scale
; option says that voice number 0 steps within the scale made current by 'two-scales' 
; and voice number 1 always steps within the predefined wholetone scale. Two 
; step-start values are given at the start of the algorithm, the absolute pitches vary 
; within the range of the scales.


;;;==========================================;;;


; Ex. 9b

(pat-events 
 (loop repeat 2 collect
       (pat random
            (pat cycle #(t) :for (pat pval (between 5 20)))
            (pat cycle t #(t) :for (pat pval (between 5 20)))
            (pat cycle t. #(t.) :for (pat pval (between 5 20)))))
 (loop repeat 2 collect (pat random 1 -1 2 -2))
 '(10 5)
 :dyn (loop repeat 2 collect (pat heap 70 90 110)) 
 :v-rhythm-pats (loop repeat 2 collect (pat cycle 2))
 :v-data-pats (list
               (pat cycle :of `(,scale-1a ,scale-1b) :name 'two-scales)
               (pat cycle 35 80 :name 'start-there))
 :abs-times t
 :max-time 30
 :tempo 180
 :step-start (list
              (pat cycle 
                   (pat cycle 35 :for 5)
                   (pat cycle 80 :for 5)
                   :for 1)
              'start-there) 
 :pitch-scale `(two-scales ,scale-2)
 :no-8 t
 :no-1 t)


; In Ex. 9b there is the same treating of pitch-scales as in 9a, but a second v-data 
; pattern named 'start-there' switches between a low and a high value every two
; seconds. For voice number 0 a step-start pattern is defined directly, so every fifth
; event starts from an alternating register. For voice number 1 the step-start option
; refers to 'start-there', changing the register (for a possible but not necessary entry) 
; every two seconds. The periodicity of 'start-there' doesn't play a role. As 
; :strict-step-start is not set the step-start values are just a base for the addition of the 
; step values, moreover they do not necessarily lie within the scales.


;;;==========================================;;;


; Ex. 10

(install-setup #!temp)

(setf v-data
      (let ((key (pat random 
                      (pat cycle 60 
                           :returning #'(lambda (x) (+ x (pick 0 12 24 36))))
                      ((pat pval (between 80 100)) weight 0.1))))
        (list
         (pat pval (new midimsg
                        time (now)
                        msg (make-program-change 3 (between 0 15))))
         (pat pval (new midimsg
                        time (now)
                        msg (make-control-change 3 10 (pick 0 63 127))))
         (pat pval (new midi time (now)
                        duration 0.2
                        amplitude (between 80 110)
                        channel 3
                        keynum (next key))))))

(setf v-rhy
      (list
       (pat cycle t)
       (pat cycle t)
       (pat cycle t)))

(pat-events nil nil nil
            :v-data-pats v-data
            :v-rhythm-pats v-rhy
            :tempo 90)

; Three virtual voices, which produce program changes, panning messages and midi objects.
; Wired instrumental voices do not occur, so there are no pitch patterns, no
; rhythm patterns and no instruments in 'pat-events'. The keynumbers of the 
; midi objects are read from a pattern defined in a lexical context.

(send-equal-pan)

; reset pan


;;;==========================================;;;


; The variations of the following example show the different
; modes of case and cond pattern. Here it's always a case
; pattern. See the reference for an example of the more general
; cond pattern.

  
; Ex. 11a
                    
(install-setup #!qt)

(setf rhy 
      (list 
       (pat random (pat cycle s. :for 4) h w q.)
       (pat cycle 
            (pat random (pat cycle x #(x) x #(x))
                 (pat cycle t t) :for 1)
            (pat random s 
                 (pat cycle t #(t)) :for 2))))

(setf mel 
      (list 
       (pat pval (between 20 32)
            :name 'a
            :returning #'(lambda (x) (list x (+ x 19.5))))
       (pat case
            :on 'a
            :type :item
            :by (t (pat line 
                        (pat chord
                             (pat pval (+ (pick 57.5 57) (car (pv 'a))))
                             (pat pval (+ (pick 60.5 61) (car (pv 'a))))
                             :for (pat random 3 5 11))
                        (pat chord
                             (pat pval (+ (between 25 70) (car (pv 'a))))
                             (pat pval (+ (between 25 70) 0.5 (car (pv 'a))))))))))

(setf dyn 
      (list 
       (pat random 70 90 110)
       (pat range :from 85 downto 80 to 105
            stepping (pat random -5 5))))
      
(pat-events rhy mel 3 
            :dyn dyn 
            :tempo 65)

; Example for the item-modus in two voices: The case pattern has no
; case distinction, just the catching condition t,
; but it refers to "a", the pattern of the bass voice
; and takes notice of every nexting of it.
; At the next 'next'-call of the case pattern it generates a new line pattern, 
; starting with a (deformed) repetition depending on the the current 
; last value of "a" and ending up with random pitches.


;;;==========================================;;;


; Ex. 11b

(install-setup #!qt)

(setf rhy 
      (list 
       (pat heap h w q.
            (pat cycle s. :for 4) 
            :name 'b)
       (pat cycle 
            (pat random (pat cycle x #(x) x #(x))
                 (pat cycle t t) :for 1)
            (pat random s 
                 (pat cycle t #(t)) :for 2))))

(setf mel 
      (list 
       (pat pval (between 20 32)
            :name 'a
            :returning #'(lambda (x) (list x (+ x 19.5))))
       (pat case
            :on 'b
            :type :condition
            :by (s. (pat range :from (pat pval (between 95 100))
                         downto 30 
                         returning #'(lambda (x) (list x (+ x 0.5) (+ x 1))))
                    t (pat line 
                           (pat chord
                                (pat pval (+ (pick 57.5 57) (car (pv 'a))))
                                (pat pval (+ (pick 60.5 61) (car (pv 'a))))
                                :for (pat random 3 5 11))
                           (pat chord
                                (pat pval (+ (between 25 70) (car (pv 'a))))
                                (pat pval (+ (between 25 70) 0.5 (car (pv 'a))))))))))
 
(setf dyn 
      (list 
       (pat random 70 90 110)
       (pat range :from 85 downto 80 to 105
            stepping (pat random -5 5))))
      
(pat-events rhy mel 3 
            :dyn dyn 
            :tempo 65)



; Example for the condition-modus in two voices: 
; There are two conditions in the case pattern:
; The entry of a dotted sixteenth of rhythm pattern "b" 
; causes the case (after 'next') to start with a new 
; descending range-pattern. Other values lead to the 
; "catching" second condition: again a new line-pattern 
; is generated and starts with repetitions, but now the 
; repetitions start only after a condition change or at 
; the beginning of the case. 
; It should be said, that pattern "a" is used in the line-patterns,
; but now it's not a reference pattern in the sense
; that its nexting influences the case distinction behaviour
; of the defined type.


;;;==========================================;;;


; Ex. 11c

(install-setup #!qt)

(setf rhy 
      (list 
       (pat heap h w q.
            (pat cycle s. :for 4) 
            :name 'b)
       (pat cycle 
            (pat random (pat cycle x #(x) x #(x))
                 (pat cycle t t) :for 1)
            (pat random s 
                 (pat cycle t #(t)) :for 2))))

(setf mel 
      (list 
       (pat pval (between 20 32)
            :name 'a
            :returning #'(lambda (x) (list x (+ x 19.5))))
       (pat case
            :on 'b
            :type :once
            :by (s. (pat range :from (pat pval (between 95 100))
                         downto 30 
                         returning #'(lambda (x) (list x (+ x 0.5) (+ x 1))))
                    t (pat line 
                           (pat chord
                                (pat pval (+ (pick 57.5 57) (car (pv 'a))))
                                (pat pval (+ (pick 60.5 61) (car (pv 'a))))
                                :for (pat random 3 5 11))
                           (pat chord
                                (pat pval (+ (between 25 70) (car (pv 'a))))
                                (pat pval (+ (between 25 70) 0.5 (car (pv 'a))))))))))

(setf dyn 
      (list 
       (pat random 70 90 110)
       (pat range :from 85 downto 80 to 105
            stepping (pat random -5 5))))

(pat-events rhy mel 3 
            :dyn dyn 
            :tempo 65)


; The same example in the once-modus: 
; When the first condition matches for the first time, 
; the range pattern starts in the high register, at later
; matches it continues from its state (it later starts from the high
; register again, but that's the behaviour of the range pattern, no 
; new instance of it will be produced).
; The same happens with the second condition:
; Repetitions only occur at the beginning
; of the example, later matches 
; just let continue the line-pattern. 


;;;==========================================;;;
