
;;; =============================================================== ;;;
;;; 
;;; Copyright (C) 2004 Daniel Mayer (post.dm@gmx.at) 
;;; 
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or (at your option) any later version.
;;; 
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;; 
;;; =============================================================== ;;;

;;; Gentlecoord 1.0
;;; Last Revision: 16.6.04


;;; Filter pattern


(defclass p-filter (pattern) 
  ((by :initarg by
       :initarg :by
       :initform #'identity)
   (limit :initarg limit
          :initarg :limit
          :initform nil)
   (error-handler :initarg error-handler
                  :initarg :error-handler
                  :initform #'(lambda nil (error "trial-number of filter exceeded")))))



(defmacro make-filter-pat (&rest args)     
   (let ((obj (gensym))
          filter-only-args
          filter-only-pairs
          pattern-args
          period-length
          (filter-keys '(:on :of :by :limit :error-handler :for)))
     (multiple-value-setq (pattern-args filter-only-args)
         (divide-listpairs args filter-keys #'symbol-name-eql?))
     (setf filter-only-pairs (list->alist filter-only-args)
           period-length (or (cdr (assoc :for filter-only-pairs :test #'symbol-name-eql?)) 1))
     `(let ((,obj (make-instance 'p-filter :for ,period-length ., (expand-pattern-args pattern-args))))
        (loop with args = (list ,. (mapcar #'(lambda (x) `(cons ,(quote-if-necessary (car x)) ,(cdr x)) )
                                           filter-only-pairs))
              for key in '(by limit error-handler) do
              (setf (slot-value ,obj key)
                    (or (cdr (assoc key args :test #'symbol-name-eql?))
                        (slot-value ,obj key)))
              finally (setf (slot-value ,obj 'data)
                            (or (cdr (assoc :on args :test #'symbol-name-eql?))
                                (cdr (assoc :of args :test #'symbol-name-eql?)))))
        (if (not (pattern? (slot-value ,obj 'data)))
            (input-error "filter expects one single pattern as item"))
        ,obj)))


; Expansion and next-item method for filter pattern


(defmethod next-in-pattern ((obj p-filter))
  (loop with next-item
        for i from 1 do
        (if (and (numberp (slot-value obj 'limit))
                 (= i (+ 1 (slot-value obj 'limit))))
            (funcall (slot-value obj 'error-handler)))
        (setf next-item 
              (funcall (slot-value obj 'by) 
                       (next (slot-value obj 'data))))
        (if next-item
            (return next-item))))


(setf (get 'filter ':new-expander)
      #'(lambda (inits) 
          `(make-filter-pat ,@inits)))



;;; Utilities for modifying a nested filter pattern


(defmacro make-mpd-b-defs (type &rest args)
  `(progn
    ,. (mapcar #'(lambda (x)
                  `(defmethod map-pattern-data-b (function (object ,x))
                      ,(cond ((= type 0) 
                              '(map-pattern-data function object))
                             ((= type 1)
                              '(map nil function (slot-value object 'data)))
                             ((= type 2)
                              '(declare (ignore function object)))
                             ((= type 3)
                              '(funcall function (slot-value object 'data))))))
             args)))


(make-mpd-b-defs 0 cycle random graph accumulation funcall rewrite)

(if-cm-version>= (2 4)
    (make-mpd-b-defs 0 palindrome)
    (make-mpd-b-defs 1 line))

(make-mpd-b-defs 1 chord)
(make-mpd-b-defs 2 markov range transposer)

(if-cm-version>= (2 4)
    (make-mpd-b-defs 2 join))

(make-mpd-b-defs 3 p-filter)



(defun collect-subpatterns (pat type collector)   
  (if (eql (class-name (class-of pat)) type)
      (push pat collector))
  (map-pattern-data-b  
   #'(lambda (x)
       (if (and (pattern? x)
                (not (eql (class-name (class-of x)) 'pval)))
           (setf collector                       
                 (collect-subpatterns x type collector))
           x)) pat)
  collector)


       
(defun replace-obj-slots (obj-list slot value)  
 (mapcar #'(lambda (x) (setf (slot-value x slot) value)) 
         obj-list))



;;; global-restriction-on-filter
;;; takes a filter pattern with possible
;;; filter subpatterns. For a global restriction
;;; it should be avoided to test one item twice.
;;; So the restriction slots of the subfilters 
;;; are set to closures that look if the test function
;;; has already been applied. 



(defun global-restriction-on-filter (pat fn limit error-handler)  
  (let (funcall-flag real-subfilters subfilters)
    (setf real-subfilters (collect-subpatterns (slot-value pat 'data) 
                                               'p-filter real-subfilters))
    (loop for real-subfilter in real-subfilters do
          (let ((f (slot-value real-subfilter 'by)))
            (setf (slot-value real-subfilter 'by)
                   #'(lambda (x)
                       (and (funcall f x)
                            (let ((result x))            
                                 (when (null funcall-flag)
                                       (setf result (funcall fn x))
                                       (if result 
                                           (setf funcall-flag t)))
                                 result))))))
            
    (setf (slot-value pat 'by) 
          #'(lambda (x)                           
              (let ((result x))
                (when (null funcall-flag)
                      (setf result (funcall fn x)))
                (setf funcall-flag nil)
                result)))
    (setf subfilters (cons pat real-subfilters))
    (replace-obj-slots subfilters 'limit limit)
    (replace-obj-slots subfilters 'error-handler error-handler)))
