
;;; =============================================================== ;;;
;;; 
;;; Copyright (C) 2004 Daniel Mayer (post.dm@gmx.at) 
;;; 
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or (at your option) any later version.
;;; 
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;; 
;;; =============================================================== ;;;

;;; Gentlecoord 1.0
;;; Last Revision: 16.6.04


;;; Crescendo / Diminuendo


(defun integerize-amp (amp)
  (cond ((floatp amp)
         (if (or (> amp 1)
                 (< amp 0))
             (error "amp as float must be >= 0 and <= 1"))
         (floor (* amp 127)))
        ((integerp amp)
         (if (or (> amp 127)
                 (< amp 0))
             (error "amp as integer must be >= 0 and <= 127"))
         amp)
        (t (error "amp data must be integer or float"))))



(defun parse-dyn-array (amp-vec &key (with-error t))   ; gives amp-vec, dur-relation-list and base-list 
  (let ((len (length amp-vec))
        dur-relation-list
        base
        no-amp-index
        correct-syntax?
        (err-msg (concatenate 'string "~2%Incoherent amp-vec, must contain at least two numbers ~%"
                              "(all non-negative), optionally followed by a list, indicating duration relations~%"
                              "(either a list of (amp-item-num - 1) positive numbers or nil,~%"
                              "indicating equal division of lengths), optionally followed by~%"
                              "a positive number, indicating the interpolation base or~%"
                              "a list of (amp-item-num - 1) positive numbers,~%"
                              "indicating the interpolation bases for the sections~2%")))
    (setf no-amp-index
          (position-if #'(lambda (x) (or (not (numberp x)) (minusp x))) amp-vec )
          dur-relation-list 
          (if no-amp-index (aref amp-vec no-amp-index))
          base 
          (if (and no-amp-index
                   (< (1+ no-amp-index) len))
              (aref amp-vec (1+ no-amp-index))))
    (setf correct-syntax?
          (if (or (< len 2)
                  (and no-amp-index
                       (> len (+ no-amp-index 2)))
                  (and dur-relation-list
                       (not (and (listp dur-relation-list)
                                 (every #'numberp dur-relation-list)
                                 (every #'plusp dur-relation-list)
                                 (= (length dur-relation-list) (1- no-amp-index)))))
                  (and base
                       (not (or (and (numberp base)
                                     (plusp base))
                                (and (listp base)
                                     (every #'numberp base)
                                     (every #'plusp base)
                                     (= (length base) (1- no-amp-index)))))))
              (progn
                (format t err-msg)
                (if with-error (error "")))
              t))
    (values (subseq amp-vec 0 (or no-amp-index len)) dur-relation-list base correct-syntax?)))


(defprocess set-channel-volume (chan amp)
  (output (new midimsg time (now)
               msg (make-control-change chan 7 amp))))


(defun amp-value (amp level)
  (cond ((numberp amp) 
         amp)  
        ((arrayp amp)
         level)
        (t (error "wrong amp-data, must be number or vector of numbers with optional specification list"))))
       

(defprocess free-all-channels ()      
  (process for ch in (range-list 15) do
           (output (new midimsg time (now)
                        msg (make-control-change ch 7 127)))))    
      

(defprocess set-inst-levels (inst-list level-or-list)      
  (process with level-list = (listify level-or-list)
           for inst in inst-list do
           (loop for ch in (nth inst (inst-ch-list  *current-inst-midi-setup*)) do
                 (output (new midimsg time (now)
                              msg (make-control-change ch 7 (car level-list)))))
           (if (listp level-or-list)
               (setf level-list (cdr level-list)))))    
  

(defun prepare-duration (dur level-change-delay level-change-break level-change-shrink)
    (let ((corrected-dur (* level-change-shrink dur))
          by-level-change-buffer-sum
          rest-dur)
          (if (< (+ level-change-delay
                    level-change-break
                    corrected-dur)
                 dur)
              (values level-change-delay (- dur level-change-delay level-change-break)
                      level-change-break)
              (progn
                (setf rest-dur (- dur corrected-dur)
                      by-level-change-buffer-sum
                      (/ 1 (+ level-change-delay level-change-break)))
                (values (* level-change-delay by-level-change-buffer-sum rest-dur)
                        corrected-dur
                        (* level-change-break by-level-change-buffer-sum rest-dur))))))


(defun make-dynamic-interpl-data (dur amp-vec by-level-change-step-time min-level-change-steps)
   (let (dur-relation-list 
         time-section 
         item-base 
         int-amp-vec 
         step-num
         step-time-num-list 
         interpl-data-list 
         interpl-time 
         (interpl-time-sum 0))
    (multiple-value-setq (int-amp-vec dur-relation-list item-base)
      (parse-dyn-array amp-vec))
    (setf int-amp-vec (map 'vector #'integerize-amp int-amp-vec)
          interpl-data-list (list (aref int-amp-vec 0) 0)
          time-section (float (/ dur (if dur-relation-list
                                         (apply #'+ dur-relation-list)
                                         (1- (length int-amp-vec))))))
    (loop with maybe-data = dur-relation-list
          for i from 1 below (length int-amp-vec) do
          (setf interpl-time (if maybe-data 
                                 (* (car maybe-data) time-section)
                                 time-section)
                interpl-time-sum (+ interpl-time-sum interpl-time)
                step-num (ceiling (* by-level-change-step-time interpl-time))
                step-num (if (< step-num min-level-change-steps)
                             min-level-change-steps 
                             step-num)
                maybe-data (cdr maybe-data))
          (push (/ interpl-time step-num) step-time-num-list)
          (push step-num step-time-num-list)
          (push interpl-time-sum interpl-data-list)
          (push (aref int-amp-vec i) interpl-data-list))
    (values (nreverse interpl-data-list) (nreverse step-time-num-list) item-base)))


(defprocess inst-amp-change (inst interpl-data-list step-time-num-list base)
  (process with step-counter = (cadr step-time-num-list)
           with step-num-stack = step-time-num-list
           with wait-time = (car step-time-num-list)
           with maybe-base-list = base
           for k from 0
           while (or (> step-counter 0)
                     (when step-num-stack
                           (if (and (listp maybe-base-list)
                                    (cadr maybe-base-list))
                               (pop maybe-base-list))
                           (setf step-num-stack (cddr step-num-stack))
                           (if step-num-stack 
                               (setf step-counter (cadr step-num-stack)
                                     wait-time (car step-num-stack))
                               (setf wait-time 0)))) 
           do
           (decf step-counter)
           (loop for ch in (nth inst (inst-ch-list  *current-inst-midi-setup*)) do
                 (output (new midimsg time (now) msg 
                              (make-control-change ch 7 
                                 (round (interpl (now) interpl-data-list 
                                                 :base (if (listp maybe-base-list) 
                                                           (car maybe-base-list)
                                                           base)))))))
           (wait wait-time)))


;;; Error messages and protocol


(defvar *conflict-count*)
(defvar *error-cases-history* nil)
(defvar *error-voice-number* nil)
(defvar *error-max-next* nil)


(defun conflict-cases (case-list &optional (destination t))
  (loop for i in case-list do
        (format destination 
                (concatenate 'string "~%conflict type ~3A: "
                             (case i
                               (1 "pitch exceeds lower bound")
                               (2 "pitch exceeds upper bound")
                               (3 "repetition of single pitch within specified area")
                               (4 "repetition of pitch in a chord within specified area")
                               (5 "octave step of single pitch within specified area")
                               (6 "octave step of pitch in a chord within specified area")
                               (7 "primes within chord")
                               (8 "octaves within chord")
                               (9 "primes: some key not free")
                               (10 "octaves: some pitch-class not free")
                               (11 "pitch-step greater than max-step"))) 
                i))
  (format destination "~2%"))  


(defun unsolvable-condition-message (&optional (d t))
  (format d "~2%Max-next exceeded by voice nr. ~A (in order of input list) ~%" *error-voice-number*)
  (format d "Pitch conditions could not be fullfilled within ~A go-next-steps." *error-max-next*)
  (format d "~2%Error history : ~2%~A~%" (reverse *error-cases-history*))
  (conflict-cases  (range-list 11 1) d)
  (format d "~3%Hint: Change properties of pitch-pattern or set less pitch restrictions, ")
  (format d "~%      i.e. allow octaves or primes, allow pitch repetitions or octave steps for one or more voices,")
  (format d "~%      set higher max-next, widen instrument bounds or reduce ")
  (format d "~%      the density of chords in voices with avoid options. ~2%"))


(defun conflict-protocol (voice-queue &optional (destination t))
  (let (conflict-types)
    (format destination "~3%Number of solved conflicts per voice with type specification:~%")
    (loop with voice-conflict-num 
          for v in voice-queue 
          for j from 0 do
          (setf voice-conflict-num 0)
          (format destination "~%voice ~A: " v)
          (loop with cnum 
                for k from 0 to 11 do
                (setf cnum (aref *conflict-count* v k))
                (when (/= 0 cnum)
                      (pushnew k conflict-types)
                      (format destination "(~A: ~5@A) " k cnum))
                (incf voice-conflict-num cnum))
          (format destination " total: ~10A" voice-conflict-num))
    (format destination "~%")
    (conflict-cases (sort conflict-types #'<=) destination)))


;;; Streams


(defun output-streamify (thing)
  (if (stringp thing)
      (open thing 
            :direction :output 
            :if-exists :supersede
            :if-does-not-exist :create)
      thing))
  

(defun map-or-call (maybe-list function)
  (if (and (listp maybe-list)
           (not (null maybe-list)))
      (mapcar function maybe-list)
      (funcall function maybe-list)))


(defun make-open-print-streams (print-maybe-list)
  (if (and (not (listp print-maybe-list))
           (not (null print-maybe-list)))
      (let ((result (output-streamify print-maybe-list)))
        (values result (if (streamp result) 
                           (list result))))
      (loop with stream-list
            with all-streams-list
            for thing in print-maybe-list 
            for i from 0 do
            (if (not (stringp thing))
                (push thing stream-list)
                (let* ((pos (position thing (subseq print-maybe-list 0 i) 
                                      :test #'equalp))
                       (new-item (if pos
                                     (car (last stream-list (1+ pos)))
                                     (output-streamify thing))))
                  (push new-item stream-list)
                  (if (null pos) 
                      (push new-item all-streams-list))))
            finally (return (values (nreverse stream-list)
                                    (nreverse all-streams-list))))))


(defun combine-make-open-print-streams (&rest print-maybe-lists)
  (let ((listcond (mapcar #'(lambda (x) 
                              (if (and (listp x)
                                       (not (null x)))
                                  t nil))
                          print-maybe-lists))
        streams 
        all-streams)
    (multiple-value-setq 
      (streams all-streams)
      (make-open-print-streams 
                  (apply #'append (mapcar #'(lambda (x) (strong-listify x))
                                          print-maybe-lists))))
    (loop with start-index = 0
          with result-streams
          for thing in print-maybe-lists 
          for listp in listcond do
          (if listp
              (progn
                (push (subseq streams start-index (+ start-index (length thing)))
                      result-streams)
                (incf start-index (length thing)))
              (progn
                (push (nth start-index streams) result-streams)
                (incf start-index)))
          finally 
           (return (values-list (nreverse (cons all-streams result-streams)))))))
          
             
(defun close-print-streams (stream-maybe-list)
  (map-or-call stream-maybe-list 
               #'(lambda (x) (if (streamp x) 
                                 (close x)
                                 x))))

