
;;; =============================================================== ;;;
;;; 
;;; Copyright (C) 2004 Daniel Mayer (post.dm@gmx.at) 
;;; 
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or (at your option) any later version.
;;; 
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;; 
;;; =============================================================== ;;;

;;; Gentlecoord 1.0
;;; Last Revision: 16.6.04


;;; Pattern utilities


(defun condition-tripletp (triplet)
  (and (listp triplet)
       (= (length triplet) 3)
       (and (listp (car triplet))
            (or (and (null (cdar triplet))
                     (eql (caar triplet) t))
                (every #'stringp (car triplet))))
       (functionp (cadr triplet))))


(defun pat-name (pat-ref)
  (typecase pat-ref
    (symbol (symbol-name pat-ref))
    (string pat-ref)
    (error "input must be symbol or string")))


(defun pv (pat-ref &optional (nad-warning t))
  (let* ((pat-name (pat-name pat-ref))
         (pat (find-object pat-name nil 'pattern)))
    (if (null pat)
        (error "no pattern with name: ~%~A" pat-name)
        (progn
          (if (and nad-warning
                   (eql (pattern-last-value pat) +nad+))
              (format t  "~2%Warning: Last pattern value :not-a-datum demanded from pattern ~A~2%" pat))
          (pattern-last-value pat)))))


(defun car-pv (pat)
 (let ((pv (pv pat)))
  (if (consp pv)
      (car pv)
      pv)))


;;; Input macros


(defmacro ever ()
  `(lambda (&rest args) (declare (ignore args)) t))



(defmacro make-cond-triplet (lst)
  (let ((x (gensym)))
   `(let ((,x ,(if (and (listp (car lst))
                        (not (eql 'quote (caar lst))))
                   `(list ,. (car lst))
                   (car lst))))
     (list (if (eql t ,x)
               (listify ,x)
               (mapcar #'pat-name (listify ,x))) 
           ,(second lst)
           #'(lambda () ,(caddr lst))))))


;;; Cond pattern


(defclass cond (pattern) ((active-pat :initarg active-pat 
                                     :initarg :active-pat 
                                     :initarg start-pat 
                                     :initarg :start-pat 
                                     :initform nil)
                         (last-cond-index :initarg last-cond-index
                                          :initarg :last-cond-index
                                          :initform nil)
                         (ref-pat-call :initarg ref-pat-call
                                       :initarg :ref-pat-call
                                       :initform nil)
                         (renew-type :initarg type :initarg :type
                                     :initarg renew-type :initarg :renew-type
                                     :initform :once)))



(defvar *cond-links*)
(setf *cond-links* (make-hash-table :test #'equalp))


(defun link-pats (ref-pat-name related-pat)
  (declare (type string ref-pat-name))
  (if (null (gethash ref-pat-name *cond-links*))
      (setf (gethash ref-pat-name *cond-links*) 
            (list related-pat))
      (progn
        (if (object-name related-pat)
            (when (member (object-name related-pat) 
                          (gethash ref-pat-name *cond-links*)
                          :test #'equalp :key #'object-name)
                  (setf (gethash ref-pat-name *cond-links*)
                        (remove-if #'(lambda (x) (equalp x (object-name related-pat)))
                                   (gethash ref-pat-name *cond-links*)
                                   :key #'object-name))
                  (format t (concatenate 'string "~2%WARNING: old pattern with name ~S " 
                                         "deleted from global hashtable *cond-links*~%")
                          (object-name related-pat))))
        (push related-pat (gethash ref-pat-name *cond-links*)))))


; After method for 'next': lookup if there are related (cond-linked) patterns


(if-cm-version>= (2 4)  
    
    (defmethod next-1 :after ((p pattern))
      (if (object-name p)
          (loop for linked-pat in (gethash (object-name p) *cond-links*) do
                (setf (slot-value linked-pat 'ref-pat-call) t))))

    (defmethod next :after ((p pattern) &optional num)             
      (declare (ignore num))
      (if (object-name p)
          (loop for linked-pat in (gethash (object-name p) *cond-links*) do
                (setf (slot-value linked-pat 'ref-pat-call) t)))))


(defun show-conds ()
  (with-hash-table-iterator (next-cond *cond-links*)
    (loop with keys-leftp
          until (not (car (setf keys-leftp (multiple-value-list (next-cond))))) do
          (format t "~%reference ~S: ~A" (cadr keys-leftp) (caddr keys-leftp)))))


(defun clear-conds ()
  (clrhash *cond-links*))



(defmacro finish-cond (cond start by type)    ; adding specific cond slots
  (let ((condition-triplets (gensym))
        (pat-names (gensym)))
   `(let ((,condition-triplets 
           (list ,. (mapcar #'(lambda (x) `(make-cond-triplet ,x))
                           (make-tripletlist by))))  
          (,pat-names))
  
      (cond ((not (every #'condition-tripletp ,condition-triplets))
             (input-error "conditions defined by :by must be list of "
                          "{ patternname | list of patternnames / predicate "
                          "/ patterndefinition } - triplets"))
            ((not (member ,type '(nil :item :condition :once)))
             (input-error "type must be one of :item :condition :once")))
      (setf ,pat-names
            (remove t (remove-duplicates 
                       (reduce #'union (mapcar #'car ,condition-triplets)) :test #'equalp))
            (slot-value ,cond 'data) 
            ,condition-triplets
            (slot-value ,cond 'active-pat)
            ,start
            (slot-value ,cond 'renew-type)
            (or ,type :once))
      (if (not (or (null (slot-value ,cond 'active-pat))
                   (pattern? (slot-value ,cond 'active-pat))))
          (input-error "start must be pattern or nil"))
      (loop for pat-name in ,pat-names do
            (link-pats pat-name ,cond))
      ,cond)))



(defmacro make-cond-pat (&rest args)     
      (let ((obj (gensym))
            pattern-args 
            cond-args 
            cond-pairs)
        (multiple-value-setq     ; separates cond args from other pattern args 
          (pattern-args cond-args) 
          (divide-listpairs args '(:start :by :type) #'symbol-name-eql?))
        (setf cond-pairs (list->alist cond-args))
        `(let ((,obj))
           (setf ,obj                    ; pseudo-data necessary !
                 (make-instance 'cond :of 'x ,. (expand-pattern-args pattern-args)))
           (finish-cond ,obj 
                        ,(cdr (assoc :start cond-pairs :test #'symbol-name-eql?))
                        ,(cdr (assoc :by cond-pairs :test #'symbol-name-eql?))
                        ,(cdr (assoc :type cond-pairs :test #'symbol-name-eql?))))))


; Expansion and next-item method for cond pattern


(setf (get 'cond ':new-expander)
      #'(lambda (inits) 
          `(make-cond-pat ,@inits)))


(defmethod next-in-pattern ((obj cond))        
  (let ((condition-triplets (slot-value obj 'data))
        next-cond-index
        pat-or-patfn)
     (setf next-cond-index
           (if (or (slot-value obj 'ref-pat-call)
                   (null (slot-value obj 'active-pat)))
               (position-if #'(lambda (x) (apply (cadr x) (car x)))
                            condition-triplets)
               (slot-value obj 'last-cond-index)))
     (if (null next-cond-index)
         (if (null (slot-value obj 'active-pat)) 
             (error "no condition true, no active pattern - cond called too early?")
             (progn 
               (setf (slot-value obj 'ref-pat-call) nil)
               (next (slot-value obj 'active-pat))))
         (progn
           (when (or (and (member (slot-value obj 'renew-type) '(:condition :once))  
                          (or (null (slot-value obj 'active-pat))
                              (not (eql (slot-value obj 'last-cond-index)
                                        next-cond-index))))
                     (and (eql (slot-value obj 'renew-type) :item)  
                          (or (null (slot-value obj 'active-pat))
                              (slot-value obj 'ref-pat-call))))
                 (setf pat-or-patfn (caddr (nth next-cond-index condition-triplets)))
                 (setf (slot-value obj 'active-pat)
                       (if (pattern? pat-or-patfn)
                           pat-or-patfn  ; if type :once the pattern itself is stored
                           (let ((must-be-pat (funcall pat-or-patfn)))
                             (if (not (pattern? must-be-pat))
                                 (error "wrong cond definition, pattern expected as third value of triplet")
                                 must-be-pat)))
                       (slot-value obj 'last-cond-index)
                       next-cond-index)
                 (if (eql (slot-value obj 'renew-type) :once)
                     (setf (caddr (nth next-cond-index condition-triplets))  
                           (slot-value obj 'active-pat))))
           (setf (slot-value obj 'ref-pat-call) nil)
           (next (slot-value obj 'active-pat))))))


(defmethod map-pattern-data-b (function (object cond))
  (declare (ignore function object)))


;;;  Case pattern


(defclass case (cond) ())


(defmacro parse-pred-data (input)
  (if (listp input)
    (if (or (and (symbolp (car input))
                 (fboundp (car input)))
            (member (car input)
                    '(function lambda)))
      input
      `(list ,. input))
    input))


; Expansion of case pattern

(defmacro make-case-pat (&rest args)     
   (let (pattern-args 
         case-args
         start
         (pr (gensym)) 
         (by (gensym)) 
         (on (gensym)) 
         (pred (gensym)))
     (multiple-value-setq     ; separates case args from other pattern args 
          (pattern-args case-args) 
          (divide-listpairs args '(:on :by :pred :start) #'symbol-name-eql?))
     (setf case-args (list->alist case-args)
           on (cdr (assoc :on case-args :test #'symbol-name-eql?))
           start (cdr (assoc :start case-args :test #'symbol-name-eql?))
           by (make-pairlist (cdr (assoc :by case-args :test #'symbol-name-eql?)))
           pred (or (cdr (assoc :pred case-args :test #'symbol-name-eql?)) #'eql))
           ; predicate or list of, should take 2 arguments, the first one for the last pattern value
     `(let ((,pr (parse-pred-data ,pred)))
        (if (not (right-data? ,pr #'functionp ,(length by)))
            (input-error "pred must be form that evals to a function "
                         "(or a list of right length of functions), "
                         "or a list (of right length) of forms that all eval to functions"))
        (change-class 
         (new cond 
              :start ,start
              :by ,(loop with item
                         for cond-pair in by
                         for rest on by 
                         for i from 0 do
                         (setf item (car cond-pair))
                         append (if (and (null (cdr rest))
                                         (eql item t))
                                    (list on '(ever) (cadr cond-pair))
                                    (list on `(lambda (x) (member (pv x) 
                                                                  ',(listify item) 
                                                                  :test (if-nth ,i ,pr) )) 
                                          (cadr cond-pair))))
              ,. pattern-args)
         'case))))


(setf (get 'case ':new-expander)
      #'(lambda (inits) 
          `(make-case-pat ,@inits)))


