
;;; =============================================================== ;;;
;;; 
;;; Copyright (C) 2004 Daniel Mayer (post.dm@gmx.at) 
;;; 
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or (at your option) any later version.
;;; 
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;; 
;;; =============================================================== ;;;

;;; Gentlecoord 1.0
;;; Last Revision: 16.6.04


;;; Expansion utilities


(defun split-by (lst pred)
  (do ((head nil (push (car tail) head))
       (tail lst (cdr tail)))
      ((or (null tail)
           (funcall pred (car tail)))
       (list (reverse head) tail))))


(defvar *p-type-list*)
(setf *p-type-list*
      '(:cycle :random :line :heap :palindrome :markov :graph :funcall :accumulation
        :rotation :rewrite :range :chord :pval :cond :case :repetition
        :recursion :filter :transposer))


(if (cm-version>= '(2 4))
    (push :join *p-type-list*))
                  

(defun p-type-p (x)
 (member x *p-type-list* :test #'symbol-name-eql?))


(defun p-type-expr-p (x)
 (and (listp x)
      (symbol-name-eql? (car x) :pat)
      (member (car x) *p-type-list* :test #'symbol-name-eql?)))


;;; Item-list expansion


(defun p-item-parse (item)
   (if (listp item)
       (cond ((and (symbol-name-eql? (car item) :pat)
                   (p-type-p (cadr item)))
              (macroexpand-1 item))
             ((symbol-name-eql? (car item) :new)
              item)
             (t `(quote ,item)))
       `(quote ,item)))


(defun make-p-item-list-expander (p-item-parser)
  #'(lambda (lst) (list* 'list (mapcar p-item-parser lst))))


(defun nested-p-item-parse (item)
   (if (listp item)
       (cond ((and (symbol-name-eql? (car item) :pat)
                   (p-type-p (cadr item)))
              (macroexpand-1 item))
             ((symbol-name-eql? (car item) :new)
              item)
             (t (funcall (make-p-item-list-expander #'nested-p-item-parse) item)))
       `(quote ,item)))


(defvar *p-item-list-expanders* )
(setf *p-item-list-expanders* (make-hash-table :test 'equalp))


(loop with standard-p-item-list-expander = (make-p-item-list-expander #'p-item-parse)
      for type in '("cycle" "line" "heap" "palindrome" "markov" "accumulation"
                    "rotation" "range" "pval" "chord" "join") do
      (setf (gethash type *p-item-list-expanders*) standard-p-item-list-expander))


;;; Nonstandard item-list expansions

; Stepping down one level for random, rewrite and graph


(loop for type-string in '("random" "rewrite" "graph") do
      (setf (gethash type-string *p-item-list-expanders*)
            (make-p-item-list-expander #'nested-p-item-parse)))


(setf (gethash "funcall" *p-item-list-expanders*) 
      #'car
      (gethash "pval" *p-item-list-expanders*) 
      #'(lambda (x) `(list (new pval ,(car x))))
      (gethash "transposer" *p-item-list-expanders*) 
      #'(lambda (data)
          (let ((parsed-data (mapcar #'p-item-parse data)))
            (if (and (not (cdr parsed-data))
                     (listp (car parsed-data))
                     (symbol-name-eql? (caar parsed-data) :new))
                (car parsed-data)
                (list* 'list parsed-data))))
      (gethash "repetition" *p-item-list-expanders*) 
      #'(lambda (x) (if (cdr x)
                        (error "repetition expects one single reference pattern"))
         (funcall #'p-item-parse (car x)))
      (gethash "filter" *p-item-list-expanders*) 
      #'(lambda (x) (if (cdr x)
                        (error "filter expects one single reference pattern"))
         (funcall #'p-item-parse (car x))))


;;; Spec-list expansion


(defun p-spec-item-parse (item)
   (if (and (listp item)
            (symbol-name-eql? (car item) :pat)
            (p-type-p (cadr item)))
       (macroexpand-1 item)
       item))


(defun p-spec-list-expand (lst)
  (loop with collector
        for (x y) on lst by #'cddr do
        (push x collector)
        (push (p-spec-item-parse y) collector)
        finally (return (reverse collector))))


(defvar *p-spec-list-expanders* )
(setf *p-spec-list-expanders* (make-hash-table :test 'equalp))


(loop for type in '("cycle" "random" "line" "heap" "palindrome" "markov" "graph" "funcall"
                    "accumulation" "rotation" "rewrite" "range" "pval"  
                    "chord" "repetition" "filter" "transposer" "join") do
      (setf (gethash type *p-spec-list-expanders*) #'p-spec-list-expand))


;;; Nonstandard spec-list expansions


(loop for pat-type-name in '("cond" "case") do
      (setf (gethash pat-type-name *p-spec-list-expanders*) 
            #'(lambda (lst)
                (loop with collector
                      for (x y) on lst by #'cddr do
                      (push x collector)
                      (if (symbol-name-eql? x :by)
                        (push (mapcar #'p-spec-item-parse y) collector)
                        (push (p-spec-item-parse y) collector))
                      finally (return (reverse collector))))))



(defun spec-list-unkeywordify (spec-list avoid-keyword-list)
  (loop for tail on spec-list by #'cddr and
        key in spec-list by #'cddr
        do 
        (if (and (keywordp key)
                 (member key avoid-keyword-list
                         :test #'symbol-name-eql?))
            (setf (car tail) (read-from-string (symbol-name key))))
        finally (return spec-list)))


(defmacro pat (&rest args)
  (let* ((pat-type (car args))
         (split-args (split-by (cdr args) #'keywordp))
         (data (car split-args))
         (spec-list (cadr split-args))
         (data-expander (gethash (symbol-name pat-type) *p-item-list-expanders*))
         (spec-list-expander (gethash (symbol-name pat-type) *p-spec-list-expanders*))
         (new-data)
         (new-spec-list))
    (if (not (member pat-type *p-type-list* :test #'symbol-name-eql?))
        (error "pattern-type does not exist or is not supported by this expansion"))
    (if (symbol-name-eql? pat-type :pval)
        (setf pat-type 'cycle))
    (spec-list-unkeywordify spec-list '(:keynums :notes :amplitudes :rhythms))
    (if data
        (setf new-data (list 'of (funcall data-expander data))
              new-spec-list (funcall spec-list-expander spec-list))
        (if (and spec-list
                 (symbol-name-eql? (car spec-list) :of))
            (setf new-data (list 'of (cadr spec-list))
                  new-spec-list (funcall spec-list-expander (cddr spec-list)))
            (setf new-data nil
                  new-spec-list (funcall spec-list-expander spec-list))))
    `(new ,pat-type ,. new-data ,. new-spec-list)))


